#include "card.h"
#include "card_private.h"
#include "sort_order.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>

int less_than (struct card *a, struct card *b)
{
    return strcmp (a->title, b->title) < 0;
}

int count_cards (struct card *p)
{
    return count_cards_priv (head (p));
}

void chop_text_linefeed (char *s, int s_length, int ch)
{
    void *lf;

    while ((lf = memmem (s, s_length, TEXT_NEWLINE, TEXT_NEWLINE_LENGTH)))
        memset (lf, ch, TEXT_NEWLINE_LENGTH);
}

void foreach_card (struct card *cardfile, void (*f)(struct card *card))
{
    cardfile = head (cardfile);

    while (cardfile)
    {
        f (cardfile);
        cardfile = cardfile->next;
    }
}

int foreach_card_vp
(struct card *cardfile, int (*f)(void *p, struct card *card), void *p)
{
    int ret = 1;

    cardfile = head (cardfile);

    while (cardfile)
    {
        ret = ret && f (p, cardfile);
        cardfile = cardfile->next;
    }

    return ret;
}

struct card *parse_cardfile (FILE *f)
{
    enum cardfile_type crd_type;
    unsigned short n;
    struct card_index *cdx;
    struct card *cardfile, *card;
    int i;

    if (! determine_type (f, &crd_type) || ! get_num_cards (f, crd_type, &n) ||
        (cdx = malloc (sizeof (struct card_index) * n)) == NULL)
        return NULL;

    if (! get_card_indices (f, crd_type, cdx, n) ||
        (cardfile = alloc_blank_cardfile (n)) == NULL)
    {
        free (cdx);
        return NULL;
    }

    for (i = 0, card = cardfile; i < n; i++, card = card->next)
        if (! get_card (f, crd_type, &cdx[i], card)) 
        {
            delete_cardfile (cardfile);
            free (cdx);
            return NULL;
        }

    assert (n == count_cards (cardfile));

    free (cdx);
    return cardfile;
}

/* Writes all cards in MGC format without graphics, returns success */
int write_cardfile (FILE *f, struct card *cardfile)
{
    return write_mgc_header (f, cardfile) &&
    write_indexes (f, cardfile) &&
    foreach_card_vp (cardfile, write_one_mgc_card, f);
}

/* Pass in '\n' to insert newline, returns success, sets dirty fields */
int insert_char (struct card *card, int col, int row, char ch)
{
    if (ch == '\n')
        return insert_newline_private (card, col, row);
    else
        return insert_char_private (card, col, row, ch);
}

int insert_title_char (struct card *card, int col, char ch)
{
    assert (ch != '\n');

    return insert_title_char_private (card, col, ch);
}

/* Returns success, sets dirty fields */
int delete_at (struct card *c, int col, int row)
{
    char *p, *max;
    int length, num_del;

    p = locate_position (c, col, row);
    if (!p)
        return 0;

    max = c->text + c->text_length-1;
    num_del = p ==
        memmem (p, max - p + 1, TEXT_NEWLINE, TEXT_NEWLINE_LENGTH)
        ? TEXT_NEWLINE_LENGTH : 1;
    length = max - (p+num_del) + 1;

    /* [0 .. p-1] [p .. p+num_del-1] [p+num_del .. text_length-1] */
    memmove (p, p + num_del, length);
    memset (max-num_del+1, '%', num_del);
    c->text_length -= num_del;
    
    c->dirty = 1;
    return 1;
}

int delete_title_at (struct card *c, int col)
{
    char *p, *max;
    int length;

    assert (0 <= col && col <= c->title_length);
    if (col == c->title_length)
        return 0;

    p = c->title + col;

    max = c->title + c->title_length-1;
    length = max - p;

    /* [0 .. p-1] [p] [p+1 .. max] */
    memmove (p, p + 1, length);
    *max = '%';
    c->title_length--;
    
    c->dirty = 1;
    return 1;
}

int delete_prev (struct card *c, int *col, int *row)
{
    char *p;
    int num_del;

    p = locate_position (c, *col, *row);
    if (p - c->text + 1 < TEXT_NEWLINE_LENGTH) 
        num_del = 1;
    else
        num_del = memcmp (p - TEXT_NEWLINE_LENGTH,
            TEXT_NEWLINE, TEXT_NEWLINE_LENGTH) == 0 ?
            TEXT_NEWLINE_LENGTH : 1;

    if (delete_at (c, *col - num_del, *row))
    {
        *row = row_from_location (c, p - num_del);
        *col = col_from_location (c, p - num_del);
        return 1;
    }
    else
        return 0;
}

int delete_title_prev (struct card *c, int *col)
{
    if (*col > 0)
    {
        delete_title_at (c, --*col);
        return 1;
    }
    else
        return 0;
}

int get_dirty_flag (struct card *card, int *flag)
{
    if (! card)
        return 0;

    *flag = card->dirty;
    return 1;
}

int num_rows (struct card *card)
{
    return max_row (card) + 1;
}

int max_row (struct card *card)
{
    int i;
    char *pos, *max;

    pos = card->text;
    max = card->text + card->text_length - 1;

    for (i = 0; i < TEXT_HEIGHT; i++)
        if (! get_next_row (&pos, max))
            /* Row doesn't exist */
            return i;

    /* Too many rows - corrupt card data */
    assert (0);
}

int row_num_cols (struct card *c, int row)
{
    return row_max_col (c, row) + 1;
}

/* Not including NL */
/* Can return -1 if row is zero-width */
int row_max_col (struct card *card, int row)
{
    int i, row_len;
    char *pos, *max, *start, *end, *nl_pos;

    pos = card->text;
    max = card->text + card->text_length - 1;

    for (i = 0; i < row; i++)
        if (! get_next_row (&pos, max))
            /* Row doesn't exist */
                /* Logic error in calling procedure */
                assert (0);

    start = pos;
    end = get_next_row (&pos, max) ? pos - 1 : max;

    /* Includes NL if present */
    row_len = end - start + 1;

    if (row_len < TEXT_NEWLINE_LENGTH)
        /* No NL, so end is last character in row */
        /* Minus 1 because columns start numbering at 0 */
        return row_len - 1;

    /* Room for a NL */

    /* end is the last character of NL if there is a NL */ 
    nl_pos = end - TEXT_NEWLINE_LENGTH + 1;
    if (strncmp (nl_pos, TEXT_NEWLINE, TEXT_NEWLINE_LENGTH) == 0)
        row_len -= TEXT_NEWLINE_LENGTH;

    return row_len - 1;
}

int get_row (struct card *card, char *buffer, int *row_len, int row)
{
    int i;
    char *pos, *max, *start, *end;

    pos = card->text;
    max = card->text + card->text_length - 1;

    for (i = 0; i < row; i++)
        if (! get_next_row (&pos, max))
            /* Row doesn't exist */
            return 0;

    start = pos;
    end = get_next_row (&pos, max) ? pos - 1 : max;

    *row_len = end - start + 1;
    memcpy (buffer, start, *row_len);

    return 1;
}

void set_title (struct card *card, char *buffer, int title_length)
{
    if (title_length > TITLE_WIDTH)
        title_length = TITLE_WIDTH;

    memcpy (card->title, buffer, title_length);
    card->title_length = title_length;
}

void get_title (struct card *card, char *buffer, int *title_length)
{
    memcpy (buffer, card->title, card->title_length);
    *title_length = card->title_length;
}

/* Pointer bubble sort (usually only one card out of order) */
struct card *sort_cardfile (struct card *cardfile)
{
    struct card *h;

    do
        h = head (cardfile);
    while (sort_pass (h));

    return h;
}

int duplicate_card (struct card *card)
{
    struct card *newcard;

    if ((newcard = malloc (sizeof (struct card))) == NULL)
        return 0;

    /* No pointers in struct card */
    memcpy (newcard, card, sizeof (struct card));

    if (newcard->prev)
        newcard->prev->next = newcard;
    newcard->next = card;
    card->prev = newcard;

    return 1;
}

struct card *alloc_blank_cardfile (int n)
{
    struct card *cardfile;

    if (n-- < 1 || (cardfile = alloc_blank_card ()) == NULL)
        return NULL;

    while (n--)
    {
        if (! prepend_blank_card (cardfile))
        {
            delete_cardfile (cardfile);
            return NULL;
        }

        cardfile = cardfile->prev;
    }

    return cardfile;
}

struct card *alloc_blank_card (void)
{
    struct card *card;

    if ((card = calloc (1, sizeof (struct card))) == NULL)
        return NULL;

    return card;
}

struct card *prepend_blank_card (struct card *cardfile)
{
    cardfile = head (cardfile);

    if ((cardfile->prev = alloc_blank_card ()))
        cardfile->prev->next = cardfile;

    return cardfile->prev;
}

struct card *delete_card (struct card *card)
{
    struct card *ret;

    if (! card)
        return NULL;

    if (card->prev)
        card->prev->next = card->next;
    if (card->next)
        card->next->prev = card->prev;

    ret = card->prev ? card->prev : card->next;
    free (card);

    return ret;
}

void delete_cardfile (struct card *cardfile)
{
    while (cardfile)
        cardfile = delete_card (cardfile);        
}
