#include "card_private.h"
#include "card.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>

/* if BITS_CHAR != 8 must rewrite file i/o functions and callees */

/**************************** STATIC DECLARATIONS ****************************/

/* Used by sort_cardfile */
static void swap_cards (struct card *first, struct card *second);
static void swap_cards_AB (struct card *first, struct card *second);
static void swap_cards_A_B (struct card *first, struct card *second);

/* Converts from little-endian to native-endian, returns success */
static int read_uint8 (FILE *f, unsigned char *uc);
static int read_uint16 (FILE *f, unsigned short *us);
static int read_uint32 (FILE *f, unsigned long *ul);

/* Converts from little-endian to native-endian, returns success */
static int write_uint8 (FILE *f, unsigned char uc);
static int write_uint16 (FILE *f, unsigned short us);
static int write_uint32 (FILE *f, unsigned long ul);

/* Returns success */
static int write_byte_fill (FILE *f, unsigned char uc, int n);
static int skip_bytes (FILE *f, unsigned long n);
static int read_bytes (FILE *f, char *s, unsigned long n);
static int write_bytes (FILE *f, char *s, unsigned long n);
static int seek_to (FILE *f, unsigned long n);

static int get_has_object (FILE *f, enum cardfile_type crd_type,
                           unsigned long card_offset, int *flag);
static int skip_presentation_object (FILE *f);

/* Returns meaningless offset if there is no text, returns success */
static int get_text_offset (FILE *f, enum cardfile_type crd_type,
                            unsigned long card_offset, unsigned long *offset);

/* Returns meaningful offset even if there is no text, returns success */  
static int get_text_len_offset (FILE *f, enum cardfile_type crd_type,
                                unsigned long card_offset, unsigned long *offset);

/**************************** STATIC DECLARATIONS ****************************/

int count_cards_priv (struct card *p)
{
    return p ? 1 + count_cards_priv (p->next) : 0;
}

int row_from_location (struct card *c, char *p)
{
    int i;
    char *pos, *max;

    pos = c->text;
    max = c->text + c->text_length - 1;

    /* Allow cursor one past end of last row */
    assert (pos <= p && p <= max + 1);

    for (i = 0; i < TEXT_HEIGHT; i++)
    {
        if (get_next_row (&pos, max))
        {
            if (p < pos)
                return i;            
        }
        else
            /* Is in last row */
            return i;
    }

    assert (0);
}

int col_from_location (struct card *c, char *p)
{
    int i;
    char *last, *pos, *max;

    pos = c->text;
    max = c->text + c->text_length - 1;

    /* Allow cursor one past end of last row */
    assert (pos <= p && p <= max + 1);

    for (i = 0; i < TEXT_HEIGHT; i++)
    {
        last = pos;
        if (get_next_row (&pos, max))
        {
            if (p < pos)
                return p - last;
        }
        else
            /* Is in last row */
            return p - last;
    }

    assert (0);
}

int write_mgc_header (FILE *f, struct card *cardfile)
{
    return seek_to (f, 0) &&
    write_uint8 (f, '\x4d') &&
    write_uint8 (f, '\x47') &&
    write_uint8 (f, '\x43') &&
    write_uint16 (f, count_cards (head (cardfile)));
}

int write_indexes (FILE *f, struct card *cardfile)
{
    int offset, ret = 1;

    cardfile = head (cardfile);
    offset = 5 + 52 * count_cards (cardfile);

    while (cardfile)
    {
        ret = ret && write_one_index (f, cardfile, offset);
        offset += bytes_one_mgc_card (f, cardfile);
        cardfile = cardfile->next;
    }

    return ret;
}

int write_one_index (FILE *f, struct card *c, int offset)
{
    assert (TITLE_WIDTH == 40);

    return
    write_uint8  (f, '\x00') &&
    write_uint8  (f, '\x00') &&
    write_uint8  (f, '\x00') &&
    write_uint8  (f, '\x00') &&
    write_uint8  (f, '\x00') &&
    write_uint8  (f, '\x00') &&
    write_uint32 (f, offset) &&
    write_uint8  (f, '\x00') &&
    write_bytes (f, c->title, c->title_length) &&
    write_byte_fill (f, '\0', TITLE_WIDTH - c->title_length) &&
    write_uint8  (f, '\x00');
}

/* Only writes text-only cards */
int write_one_mgc_card (void *f, struct card *c)
{
    return
    write_uint16 (f, 0) &&
    write_uint16 (f, c->text_length) &&
    write_bytes (f, c->text, c->text_length);
}

int bytes_one_mgc_card (void *p, struct card *c)
{
    return 2 + 2 + c->text_length;
}

int insert_newline_private (struct card *c, int col, int row)
{
    char *start, *new_start;
    int length;

    if (c->text_length + TEXT_NEWLINE_LENGTH > MAX_TEXT_BYTES)
        return 0;

    if (num_rows (c) + 1 > TEXT_HEIGHT)
        return 0;

    start = locate_position (c, col, row);
    new_start = start + TEXT_NEWLINE_LENGTH;
    /* [0 .. start-1] [start .. text_length-1] */
    length = (c->text + c->text_length-1) - start + 1; 

    memmove (new_start, start, length);
    memcpy (start, TEXT_NEWLINE, TEXT_NEWLINE_LENGTH);
    c->text_length += TEXT_NEWLINE_LENGTH;

    c->dirty = 1;
    return 1;
}

/* Doesn't handle newlines */
int insert_char_private (struct card *c, int col, int row, int ch)
{
    char *start; 
    int length;
    
    if (c->text_length + 1 > MAX_TEXT_BYTES)
        return 0;
 
    if (row_num_cols (c, row) + 1 > TEXT_WIDTH)
        return 0;
    
    start = locate_position (c, col, row);   
    /* [0 .. start-1] [start .. text_length-1] */
    length = (c->text + c->text_length-1) - start + 1;
 
    memmove (start + 1, start, length);
    *start = ch;
    c->text_length++;

    c->dirty = 1;
    return 1;
}

int insert_title_char_private (struct card *c, int col, int ch)
{
    char *start; 
    int length;
    
    if (c->title_length + 1 > TITLE_WIDTH)
        return 0;
 
    start = c->title + col;
    /* [0 .. start-1] [start .. text_length-1] */
    length = (c->title + c->title_length-1) - start + 1;
 
    memmove (start + 1, start, length);
    *start = ch;
    c->title_length++;

    c->dirty = 1;
    return 1;
}

char *locate_position (struct card *c, int col, int row)
{
    int i, max_col;
    char *pos, *max, *start, *end;

    if (row < 0 || (col < 0 && row == 0))
        return NULL;
    
    pos = c->text;
    max = c->text + c->text_length - 1;
                
    for (i = 0; i < row; i++)
        if (! get_next_row (&pos, max))
                /* Row doesn't exist */
                return NULL;
 
    start = pos;
    end = get_next_row (&pos, max) ? pos - 1 : max;
    max_col = end - start;

    /* Allow position at end of row */
    if (max_col + 1 < col)
        /* Col doesn't exist */
        return NULL;

    return start + col;
}

struct card *head (struct card *p)
{
    if (p == NULL)
    	return NULL;

    while (p->prev)
    	p = p->prev;

    return p;
}

int sort_pass (struct card *c)
{
   int swapped = 0;

   while (c->next)
   {
       if (less_than (c->next, c) && (swapped = 1))
           swap_cards (c, c->next);
       else
           c = c->next;
   }

   return swapped;
}

static void swap_cards (struct card *first, struct card *second)
{
    /* If first and second are adjacent,
       have to ensure we don't get self-loops */

    if (first->next == second)
        swap_cards_AB (first, second);
    else if (second->next == first)
        swap_cards_AB (second, first);
    else
        swap_cards_A_B (first, second);
}

static void swap_cards_AB (struct card *first, struct card *second)
{
    assert (first && second);
    assert (first->next == second && first == second->prev);

    if (first->prev)
    	first->prev->next = second;
    if (second->next)
    	second->next->prev = first;

    first->next = second->next;
    second->prev = first->prev;

    first->prev = second;
    second->next = first;
}

static void swap_cards_A_B (struct card *first, struct card *second)
{
    struct card *p;

    assert (first && second);

    if (first->prev)
    	first->prev->next = second;
    if (second->prev)
    	second->prev->next = first;
    if (first->next)
    	first->next->prev = second;
    if (second->next)
    	second->next->prev = first;

    p = first->prev;
    first->prev = second->prev;
    second->prev = p;

    p = first->next;
    first->next = second->next;
    second->next = p;
}

static int write_byte_fill (FILE *f, unsigned char uc, int n)
{
    while (n--)
        if (! write_uint8 (f, uc))
            return 0;

    return 1;
}

static int read_uint8 (FILE *f, unsigned char *uc)
{
    fread (uc, 1, 1, f);

    return ! feof (f);
}

static int read_uint16 (FILE *f, unsigned short *us)
{
    unsigned char uc;

    fread (&uc, 1, 1, f);
    *us = uc;

    fread (&uc, 1, 1, f);
    *us += 256 * uc;

    return ! feof (f);
}

static int read_uint32 (FILE *f, unsigned long *ul)
{
    unsigned char uc;

    fread (&uc, 1, 1, f);
    *ul = uc;

    fread (&uc, 1, 1, f);
    *ul += 256 * uc;

    fread (&uc, 1, 1, f);
    *ul += 65536L * uc;

    fread (&uc, 1, 1, f);
    *ul += 16777216L * uc;

    return ! feof (f);
}

static int write_uint8 (FILE *f, unsigned char uc)
{
    return fwrite (&uc, 1, 1, f);
}

static int write_uint16 (FILE *f, unsigned short us)
{
    unsigned char uc;
    int written;

    uc = us % 256;
    written =  fwrite (&uc, 1, 1, f);

    uc = us / 256;
    written += fwrite (&uc, 1, 1, f);

    return written == 2;
}

static int write_uint32 (FILE *f, unsigned long ul)
{
    unsigned char uc;
    int written;

    uc =  ul % 256;
    written = fwrite (&uc, 1, 1, f);

    uc = (ul / 256) % 256;
    written += fwrite (&uc, 1, 1, f);

    uc = (ul / 65536L) % 256;
    written += fwrite (&uc, 1, 1, f);

    uc =  ul / 16777216L;
    written += fwrite (&uc, 1, 1, f);

    return written == 4;
}

static int skip_bytes (FILE *f, unsigned long n)
{
    return ! fseek (f, n, SEEK_CUR);
} 

static int read_bytes (FILE *f, char *s, unsigned long n)
{
    fread (s, 1, n, f);
    return ! feof (f);
}

static int write_bytes (FILE *f, char *s, unsigned long n)
{
    return fwrite (s, 1, n, f) == n;
}

static int seek_to (FILE *f, unsigned long n)
{
    return ! fseek (f, n, SEEK_SET);
}

int determine_type (FILE *f, enum cardfile_type *crd_type)
{
    char signature[4];

    seek_to (f, 0);
    read_bytes (f, signature, 3);
    signature[3] = '\0';

    if (strcmp (signature, "MGC") == 0)
    	*crd_type = MGC;
    else if (strcmp (signature, "RRG") == 0)
    	*crd_type = RRG;
    else if (strcmp (signature, "DKO") == 0)
    	*crd_type = DKO;
    else return 0;

    return 1;
}

int get_num_cards (FILE *f, enum cardfile_type crd_type,
                   unsigned short *n)
{
    switch (crd_type)
    {
    case RRG:
    case DKO:
    	return seek_to (f, 7) && read_uint16 (f, n);
    case MGC:
    	return seek_to (f, 3) && read_uint16 (f, n);
    default:
    	return 0;
    }
}

int get_card_indices (FILE *f, enum cardfile_type crd_type,
                      struct card_index *cdx, unsigned short n)
{
    int i;

    /* DKO not supported yet */
    if (crd_type != MGC && crd_type != RRG)
    	return 0;

    if (! seek_to (f, crd_type == MGC ? 5 : 9))
    	return 0;

    for (i = 0; i < n; i++)
    {
    	char t[TITLE_WIDTH + 1];

    	skip_bytes (f, 6);
    	read_uint32 (f, &cdx[i].card_offset);
    	read_bytes (f, &cdx[i].flag, 1);
    	read_bytes (f, cdx[i].title, TITLE_WIDTH);
    	skip_bytes (f, 1);

    	memcpy (t, cdx[i].title, TITLE_WIDTH);
    	t[TITLE_WIDTH] = '\0';
    	cdx[i].title_length = strlen (t);
    }

    return 1;
}

static int get_has_object (FILE *f, enum cardfile_type crd_type, 
                     unsigned long card_offset, int *flag)
{
    unsigned short bitmap_length;

    /* First uint16 indicates bitmap length for MGC and boolean has_object for RRG */
    if ((crd_type == MGC || crd_type == RRG || crd_type != DKO) &&
        seek_to (f, card_offset) && read_uint16 (f, &bitmap_length))
    {
    	*flag = ! ! bitmap_length;
    	return 1;
    }

    return 0;
}

static int get_text_offset (FILE *f, enum cardfile_type crd_type,
                            unsigned long card_offset, unsigned long *offset)
{
    if (crd_type == MGC)
    {
    	unsigned short bitmap_length;

    	if (! seek_to (f, card_offset) || ! read_uint16 (f, &bitmap_length))
    		return 0;

    	*offset = card_offset + (bitmap_length == 0 ? 4 : bitmap_length + 12);

    	return 1;
    }
    else if (crd_type == RRG)
    {
        unsigned short has_ole;

    	if (! seek_to (f, card_offset) || ! read_uint16 (f, &has_ole))
    		return 0;

        if (has_ole)
        {
            unsigned long ole_ver;
            unsigned long ole_format, l;

            /* Obj ID */
            if (! read_uint32 (f, &l))
                return 0;

            /* 1 for nt 3.51, 1.5 for win 3.1 */
            if (! read_uint32 (f, &ole_ver) ||
                (ole_ver != 0x0001 && ole_ver != 0x0501))
                abort ();

            if (! read_uint32 (f, &ole_format) || ! read_uint32 (f, &l) ||
                ! skip_bytes (f, l))
                return 0;

            if (ole_format == 3)
            {
                /* Static object */
                if (! read_uint32 (f, &l) || ! read_uint32 (f, &l) ||
                    ! read_uint32 (f, &l) || ! skip_bytes (f, l))
                    return 0;
            }
            else
            {
                /* Linked or embedded */
                if (! read_uint32 (f, &l) || ! skip_bytes (f, l) ||
                    ! read_uint32 (f, &l) || ! skip_bytes (f, l))
                    return 0;
                if (ole_format == 2)
                {
                    /* Embedded */
                    if (! read_uint32 (f, &l) || ! skip_bytes (f, l))
                        return 0;
                    if (! skip_presentation_object (f))
                        return 0;
                }
                else
                {
                    /* Linked */
                    if (! read_uint32 (f, &l) || ! skip_bytes (f, l) ||
                        ! read_uint32 (f, &l) || ! read_uint32 (f, &l))
                        return 0;
                    if (! skip_presentation_object (f))
                        return 0;
                }
            }

            /* Nine uint16 - char width & height, left, top, right, bottom, obj type */
            *offset = ftell (f) + 16;
        }
        else
            *offset = card_offset + 4;

    	return 1;
    }
    else
    	return 0;
}

static int skip_presentation_object (FILE *f)
{
    unsigned long l, ole_ver;
    unsigned char c;
    int i, gfx;

    /* ole version */
    if (! read_uint32 (f, &ole_ver) ||
        (ole_ver != 0x0001 && ole_ver != 0x0501))
        abort ();

    /* format */
    if (! read_uint32 (f, &l))
        return 0;

    /* No presentation object */
    if (l == 0)
        return 1;

    if (! read_uint32 (f, &l) || ! read_uint8 (f, &c))
        return 0;

    gfx = 0;

    /* Try to match METAFILEPICT\0 or BITMAP\0 or DIB\0 */
    switch (c)
    {
    case 'M':
        for (i = 0; i <= 11; i++)
        {
            if (! read_uint8 (f, &c))
                return 0;

            if (c != "ETAFILEPICT"[i])
            {
                while (c)
                    if (! read_uint8 (f, &c))
                        return 0;
                break;
            }
        }
        gfx = 1;
        break;

    case 'B':
        for (i = 0; i <= 5; i++)
        {
            if (! read_uint8 (f, &c))
                return 0;

            if (c != "ITMAP"[i])
            {
                while (c)
                    if (! read_uint8 (f, &c))
                        return 0;
                break;
            }
        }
        gfx = 1;
        break;

    case 'D':
        for (i = 0; i <= 2; i++)
        {
            if (! read_uint8 (f, &c))
                return 0;

            if (c != "IB"[i])
            {
                while (c)
                    if (! read_uint8 (f, &c))
                        return 0;
                break;
            }
        }
        gfx = 1;
        break;
    }

    if (gfx)
    {
        if (! read_uint32 (f, &l) || ! read_uint32 (f, &l) ||
            ! read_uint32 (f, &l) || ! skip_bytes (f, l))
            return 0;
    }
    else
    {
        if (! read_uint32 (f, &l))
            return 0;
        if (! l)
            if (! read_uint32 (f, &l) || ! skip_bytes (f, l))
                return 0;
        if (! read_uint32 (f, &l) || ! skip_bytes (f, l))
            return 0;
    }

    return 1;
}

static int get_text_len_offset (FILE *f, enum cardfile_type crd_type,
                                unsigned long card_offset, unsigned long *offset)
{
    /* in all three cardfile types, the text length is the uint16
       preceeding the text itself */

    if (! get_text_offset (f, crd_type, card_offset, offset))
    	return 0;

    *offset -= 2;
    return 1;
}

int get_card (FILE *f, enum cardfile_type crd_type,
              struct card_index *cdx, struct card *card)
{
    int has_object;
    unsigned long text_offset, text_len_offset;

    if (! cdx || ! card ||
        ! get_has_object      (f, crd_type, cdx->card_offset, &has_object) ||
        ! get_text_offset     (f, crd_type, cdx->card_offset, &text_offset) ||
        ! get_text_len_offset (f, crd_type, cdx->card_offset, &text_len_offset))
        return 0;

    memcpy (card->title, cdx->title, TITLE_WIDTH);
    card->title_length = cdx->title_length;

    /* We strip out the objects -- alert user before save */
    card->object_skipped = has_object;

    return seek_to (f, text_len_offset) &&
           read_uint16 (f, &card->text_length) &&
           seek_to (f, text_offset) && 
           read_bytes (f, card->text, card->text_length);
}

int get_next_row (char **pos, char *max)
{
    int search_length;
    char *newline_pos;

    search_length = min (max - *pos + 1, TEXT_WIDTH + TEXT_NEWLINE_LENGTH);
    if ((newline_pos = memmem (*pos, search_length, TEXT_NEWLINE, TEXT_NEWLINE_LENGTH)))
    {
    	*pos = newline_pos + TEXT_NEWLINE_LENGTH;
    	return 1;
    }

    if (*pos + TEXT_WIDTH > max)
    	return 0;

    *pos += TEXT_WIDTH;
    return 1;
}

void *memmem (void *haystack, int haystack_length,
              void *needle, int needle_length)
{
    if (needle_length == 0)
    	return haystack;

    while (haystack_length >= needle_length)
    {
    	if (memcmp (haystack, needle, needle_length) == 0)
    		return haystack;

    	haystack = ((char *) haystack) + 1;
    	haystack_length--;
    }

    return NULL;
}
