#include "card.h"
#include "cardfile.h"
#include "colors.h"
#include "requests.h"
#include "state.h"

#include <ctype.h>
#include <curses.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <unistd.h>

static int process_pending_requests (void);
static void handle_input (int ch);

static int is_input_pending (void);
static void init_signals (void);
static void init_curses (void);
static void inform_size (void);
static void resize (int sig);

static int is_fkey (int ch);
static int my_unctrl (int ch);


int main (int argc, char *argv[])
{
    FILE *f;
    struct card *cards;

    if ((f = fopen (argv[1], "r+b")))
    {
        cards = parse_cardfile (f);
        if (! cards)
        {
            fputs ("error parsing cardfile\n", stderr);
            return 1;
        }
    }
    else
    {
        f = fopen (argv[1], "w+b");
        if (!f)
        {
            fputs ("error opening cardfile for read and write\n", stderr);
            return 1;
        }
        cards = alloc_blank_card ();
    }

    post_request    (REQUEST_NEW_STATE);

    init_signals ();
    init_curses ();
    inform_size ();

    post_request_vp (REQUEST_INFORM_FILE, f);
    post_request_vp (REQUEST_INFORM_CARDS, cards);
    post_request    (REQUEST_REFRESH);

    while (process_pending_requests ())
    {
        int ch;

        do
        {
            errno = 0;
            ch = getch ();
            if (ch == ERR && errno == EINTR)
                /* signal interrupted */
                continue;

            handle_input (ch);
        }
        while (is_input_pending ());
    }

    endwin ();
    return 0;
}

static int is_input_pending (void)
{
    int ch;

    nodelay (stdscr, 1);

    do
    {
       errno = 0;
       ch = getch ();
    }
    while (ch == ERR && errno == EINTR);

    nodelay (stdscr, 0);
    
    if (ch == ERR)
        /* No input pending */
        return 0;

    ungetch (ch);
    return 1;
}

static void init_signals (void)
{
    signal (SIGWINCH, resize);
    siginterrupt (SIGWINCH, 1);
}

static void init_curses (void)
{
    initscr ();
    start_color ();
    keypad (stdscr, TRUE);
    raw ();
    noecho ();
    typeahead (-1);

    init_pair (BKGD_COLOR, COLOR_BLACK, COLOR_BLUE);
    init_pair (CARD_COLOR, COLOR_WHITE, COLOR_MAGENTA);
    init_pair (TEXT_COLOR, COLOR_WHITE, COLOR_MAGENTA);
    init_pair (INDEX_COLOR, COLOR_WHITE, COLOR_BLUE);
    init_pair (HILITE_COLOR, COLOR_WHITE, COLOR_YELLOW);
}

static void inform_size (void)
{
    struct winsize ws;

    /* Standard curses trick for getting it to recognize the resize */
    endwin ();
    refresh ();

    ioctl (0, TIOCGWINSZ, &ws);
    post_request_i (REQUEST_INFORM_HEIGHT, ws.ws_row);
    post_request_i (REQUEST_INFORM_WIDTH, ws.ws_col);
}

static void resize (int sig)
{
    post_request (REQUEST_RESIZE);
}

/* Returns false if quit requested */
static int process_pending_requests (void)
{
    static struct state *cur_state;

    void *vp;
    int i;

    /* Exit loop if input pending. The input may request another screen
       refresh, in which case some of our work queue should be skipped */
    while (is_request () && ! is_input_pending ())
    {
        int msg;

        msg = peek_request ();
        switch (msg)
        {
        case REQUEST_RESIZE:
            inform_size ();
            post_request (REQUEST_REFRESH);
            break;
        case REQUEST_NEW_STATE:
            cur_state = state_new ();
            break;
        case REQUEST_DUPLICATE_CARD:
            state_duplicate_card (cur_state);
            break;
        case REQUEST_INFORM_CARDS:
            vp = peek_vp ();
            state_inform_cards (cur_state, vp);
            break;
        case REQUEST_INFORM_FILE:
            vp = peek_vp ();
            state_inform_file (cur_state, vp);
            break;
        case REQUEST_INSERT:
            i = peek_i ();
            state_status_update (cur_state);
            state_insert (cur_state, i);
            break;
        case REQUEST_STEP_CARD:
            i = peek_i ();
            state_status_update (cur_state);

            /* If the screen will be repainted, don't bother right now */
            if (peek_count (REQUEST_STEP_CARD) > 1)
                state_step_card (cur_state, i, NO_REFRESH);
            else
                state_step_card (cur_state, i, REFRESH_OK);

            break;
        case REQUEST_MOVE_CURSOR:
            i = peek_i ();
            state_status_update (cur_state);
            state_move_cursor (cur_state, i);
            break;
        case REQUEST_NEW_CARD:
            state_status_update (cur_state);
            state_new_card (cur_state);
            break;
        case REQUEST_REFRESH:
            state_compute_indexes (cur_state);
            clear ();
            clearok (curscr, 1);
            state_refresh (cur_state);
            clearok (curscr, 0);
            break;
        case REQUEST_SAVE:
            state_save (cur_state);
            break;
        case REQUEST_DELETE_PREV:
            state_status_update (cur_state);
            state_delete_prev (cur_state);
            break;
        case REQUEST_DELETE_AT:
            state_status_update (cur_state);
            state_delete_at (cur_state);
            break;
        case REQUEST_INFORM_HEIGHT:
            i = peek_i ();
            state_inform_height (cur_state, i);
            break;
        case REQUEST_INFORM_WIDTH:
            i = peek_i ();
            state_inform_width (cur_state, i);
            break;
        case REQUEST_COMPUTE_INDEXES:
            state_compute_indexes (cur_state);
            break;
        case REQUEST_DELETE_CARD:
            state_status_update (cur_state);
            state_delete_card (cur_state);
            break;
        case REQUEST_GRACEFUL_EXIT:
            state_free (cur_state);
            cur_state = NULL;
            break;
        }

        pop_request ();

        if (msg == REQUEST_GRACEFUL_EXIT)
            return FALSE;
    }

    return TRUE;
}

static int is_fkey (int ch)
{
    return ch > 255;
}

static int my_unctrl (int ch)
{
    return unctrl (ch) [1];
}

static void handle_input (int ch)
{
    if (iscntrl (ch) && ! is_fkey (ch))
    {
        if (ch == 0x7f)
        {
            post_request (REQUEST_DELETE_PREV);
            return;
        }

        switch (my_unctrl (ch))
        {
        case 'L':   post_request (REQUEST_REFRESH);
                    return;

        case 'X':   post_request (REQUEST_GRACEFUL_EXIT);
                    return;

        case 'O':   post_request (REQUEST_SAVE);
                    return;

        case 'D':   post_request (REQUEST_DUPLICATE_CARD);
                    return;

        case 'N':   post_request (REQUEST_NEW_CARD);
                    return;

        case 'F':   post_request_i (REQUEST_STEP_CARD, +1);
                    return;
    
        case 'B':   post_request_i (REQUEST_STEP_CARD, -1);
                    return;

        case 'J':   post_request_i (REQUEST_INSERT, '\n');
                    return;

        case 'K':   post_request (REQUEST_DELETE_CARD);
                    return;

        default:    return;
        }
    }

    switch (ch)
    {
    case KEY_LEFT:    post_request_i (REQUEST_MOVE_CURSOR, LEFT);
            return;
    
    case KEY_UP:    post_request_i (REQUEST_MOVE_CURSOR, UP);
            return;
    
    case KEY_RIGHT:    post_request_i (REQUEST_MOVE_CURSOR, RIGHT);
            return;
    
    case KEY_DOWN:    post_request_i (REQUEST_MOVE_CURSOR, DOWN);
            return;
    
    case KEY_NPAGE:    post_request_i (REQUEST_STEP_CARD, +1);
            return;
    
    case KEY_PPAGE:    post_request_i (REQUEST_STEP_CARD, -1);
            return;

    case KEY_BACKSPACE:    post_request (REQUEST_DELETE_PREV);
            return;

    case KEY_DC:    post_request (REQUEST_DELETE_AT);
            return;

    default:    if (! is_fkey (ch)) post_request_i (REQUEST_INSERT, ch);
            return;
    }
}
