#include "card.h"
#include "cardfile.h"
#include "colors.h"
#include "state.h"
#include "utility.h"

#include <assert.h>
#include <curses.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static struct card *add_saturate (struct card *c, int n);
static struct card *subtract_saturate (struct card *c, int n);
static int subtract (struct card *a, struct card *b);
static int is_after (struct card *a, struct card *b);
static void put_at (int x, int y, int ch);
static int print_between (int x1, int y1, int x2, int y2, int ch);
static void print_at (int x, int y, char *s, int len);
static void print_at_c (int x, int y, char *s);
static int real_x (struct state *st);
static int real_y (struct state *st);

static struct card *add_saturate (struct card *c, int n)
{
    if (n < 0)
        return subtract_saturate (c, -n);

    return n && c->next ? add_saturate (c->next, n - 1) : c;
}

static struct card *subtract_saturate (struct card *c, int n)
{
    if (n < 0)
        return add_saturate (c, -n);

    return n && c->prev ? subtract_saturate (c->prev, n - 1) : c;
}

static int subtract (struct card *a, struct card *b)
{
    assert (a && b);
    return a == b ? 0 : subtract (a, b->next) + 1;
}

static int is_after (struct card *a, struct card *b)
{
    return a && (a == b || is_after (a->next, b));
}

struct state *state_new (void)
{
    struct state *s = malloc (sizeof (struct state));
    if (!s) exit(1);

    s->cur_card = NULL;
    s->file = NULL;
    s->first_index = s->last_index = NULL;
    s->status = HELP;
    s->title_mode = 0;
    return s;
}

/* Assumes first, last are both valid or NULL */
void state_compute_indexes (struct state *s)
{
    int shown_titles = s->screen_rows - 2;

    if (s->first_index == NULL)
    {
        s->first_index = s->cur_card;
        s-> last_index = add_saturate (s->cur_card, shown_titles - 1);
    }
    else
    {
        /* Scroll screen if needed */
        if (is_after (s->cur_card, s->first_index))
        {
            s->first_index = s->cur_card;
            s-> last_index = add_saturate (s->cur_card, shown_titles - 1);
        }

        if (is_after (s->last_index, s->cur_card))
        {
            s->first_index = subtract_saturate (s->cur_card, shown_titles - 1);
            s-> last_index = s->cur_card;
        }

        /* If any cards were deleted, grow on-screen window to fit */
        /* Grow down, and then grow up if needed */
        if (subtract (s->last_index, s->first_index) < shown_titles - 1)
        {
            s-> last_index =      add_saturate (s->first_index, shown_titles - 1);
            s->first_index = subtract_saturate (s->last_index,  shown_titles - 1);
        }

        /* If the screen shrunk, reduce on-screen window to fit */
        /* Trim from the bottom, then from the top if needed */
        if (subtract (s->last_index, s->first_index) > shown_titles - 1)
        {
            int top_rows = subtract (s->cur_card,  s->first_index);

            if (top_rows + 1 < shown_titles)
                /* We can trim only from the bottom */
                s-> last_index = add_saturate (s->first_index, shown_titles - 1);
            else
            {
                /* We must trim all the bottom and some from the top */
                s-> last_index = s->cur_card;
                s->first_index = subtract_saturate (s->last_index,  shown_titles - 1);
            }
        }
    }
}

void state_duplicate_card (struct state *s)
{
    assert (s->cur_card && s->first_index && s->last_index);
    duplicate_card (s->cur_card);
    state_jump_card (s, s->cur_card->prev, REFRESH_OK);
    state_compute_indexes (s);
    state_refresh (s);
}

void state_inform_cards (struct state *s, struct card *cards)
{
    assert (! s->cur_card);
    s->cur_card = cards;
    s->cursor_row = s->cursor_col = 0;
    s->title_col = 0;
    s->cur_card = sort_cardfile (s->cur_card);
    state_compute_indexes (s);

    /* Maybe the cards came from alloc_blank_card, and they weren't
       loaded from a file, but we still want to show the user how many 
       cards there are */
    s->status = LOADED;
}

void state_inform_file (struct state *s, FILE *file)
{
    assert (! s->file);
    s->file = file;
}

void state_insert (struct state *s, int ch)
{
    assert (s->cur_card && s->first_index && s->last_index);

    if (! s->title_mode)
    {
        state_move_cursor (s, STAY);
        if (insert_char (s->cur_card, s->cursor_col, s->cursor_row, ch))
            state_move_cursor (s, RIGHT);
    }
    else
    {
        if (ch == '\n')
        {
            s->title_mode = 0;
            return;
        }

        insert_title_char (s->cur_card, s->title_col, ch);
        s->title_col++;

        sort_cardfile (s->cur_card);
        state_compute_indexes (s);
    }

    state_refresh (s);
}

/* e.g. n == -5 means go five previous, n == 5 means go five forward */
void state_step_card (struct state *s, int n, int refresh_ok)
{
    assert (s->cur_card && s->first_index && s->last_index);

    if ((n > 0 && ! s->cur_card->next) ||
        (n < 0 && ! s->cur_card->prev))
        if (refresh_ok)
            state_refresh (s);

    for (; n > 0 && s->cur_card->next; n--)
        state_jump_card (s, s->cur_card->next, refresh_ok);
    for (; n < 0 && s->cur_card->prev; n++)
        state_jump_card (s, s->cur_card->prev, refresh_ok);
}

void state_jump_card (struct state *s, struct card *c, int refresh_ok)
{
    s->cur_card = c;
    s->cursor_row = s->cursor_col = 0;
    s->title_col = 0;
    state_compute_indexes (s);
    if (refresh_ok)
        state_refresh (s);
}

void state_move_cursor (struct state *s, int dir)
{
    int cc, cr;
    struct card *c;

    assert (s->cur_card && s->first_index && s->last_index);

    if (s->title_mode)
    {
        if (dir == LEFT && s->title_col)
            s->title_col--;
        if (dir == RIGHT && s->title_col < s->cur_card->title_length)
            s->title_col++;
        if (dir == DOWN)
            s->title_mode = 0;
        state_refresh (s);
        return;
    }

    cr = s->cursor_row;
    cc = s->cursor_col;
    c  = s->cur_card;

    if (dir == UP)
    {
        if (cr)
            cr--;
        else
            s->title_mode = 1;
    }
    else if (dir == DOWN)
    {
        if (cr < max_row (c))
            cr++;
    }
    else if (row_max_col (c, cr) < cc)
        if (dir == STAY)
            cc = row_max_col (c, cr) + 1;
        else if (dir == LEFT)
            if (cc == 0)
            {
                if (cr > 0)
                {
                    cr--;
                    cc = row_max_col (c, cr) + 1;
                }
            }
            else
                cc = min (row_max_col (c, cr) + 1, cc - 1);
        else
            if (++cr <= max_row (c))
                cc = 0;
            else
                cr = max_row (c), cc = row_max_col (c, cr) + 1;
    else if (dir == LEFT)
        if (cc--);
        else
            if (cr--)
                cc = row_max_col (c, cr) + 1;
            else
                cr = cc = 0;
    else if (dir == RIGHT)
        if (++cc <= row_max_col (c, cr) + 1);
        else
            if (++cr <= max_row (c))
                cc = 0;
            else
                cr = max_row (c), cc = row_max_col (c, cr) + 1;
    else;
        /* Do nothing for STAY */

    s->cursor_col = cc;
    s->cursor_row = cr;
    s->cur_card   = c;

    state_refresh (s);
}

void state_new_card (struct state *s)
{
    struct card *new_card;

    assert (s->cur_card && s->first_index && s->last_index);

    new_card = prepend_blank_card (s->cur_card);
    if (! new_card)
        exit (1);
    s->title_mode = 1;
    state_jump_card (s, new_card, REFRESH_OK);
}

void state_delete_card (struct state *s)
{
    struct card *new_card;

    assert (s->cur_card && s->first_index && s->last_index);

    /* don't delete only card */
    if (s->first_index == s->last_index)
        return;

    /* this is the card to jump to; prev if it exists, else next */
    new_card = delete_card (s->cur_card);
    if (! new_card)
        /* should not happen */
        exit (1);

    /* don't allow first_index or last_index to dangle */
    if (s->cur_card == s->first_index)
        s->first_index = new_card;
    else if (s->cur_card == s->last_index)
        s->last_index = new_card;

    state_jump_card (s, new_card, REFRESH_OK);
}

static void put_at (int x, int y, int ch)
{
    int my, mx, by, bx;

    getmaxyx (stdscr, my, mx);
    getbegyx (stdscr, by, bx);
    if (x < bx || x >= mx || y < by || y >= my)
        return; 

    move (y, x);
    addch (ch);
}   
 
/* Draws vertical or horizontal lines only */
static int print_between (int x1, int y1, int x2, int y2, int ch)
{
    int a, b;
    
    if (x1 == x2)
        for (a = min (y1, y2) + 1, b = max (y1, y2); a < b; a++)
            put_at (x1, a, ch);
    else if (y1 == y2)
        for (a = min (x1, x2) + 1, b = max (x1, x2); a < b; a++)
            put_at (a, y1, ch);
    else
        return 0;
    
    return 1;
}

static void print_at (int x, int y, char *s, int len)
{
    for (; len; len--)
        put_at (x + len-1, y, s[len-1]);
}

static void print_at_c (int x, int y, char *s)
{
    print_at (x, y, s, strlen (s));
}

static int real_x (struct state *st)
{
    return row_max_col (st->cur_card, st->cursor_row) < st->cursor_col ?
        row_max_col (st->cur_card, st->cursor_row) + 1 : st->cursor_col;
}

static int real_y (struct state *st)
{
    return st->cursor_row;
}

void state_refresh (struct state *st)
{
    int a, b, c, d, e, f, x;
    int p, q, r, s, t, u, y;
    int height;

    char title_buf[TITLE_WIDTH+1];
    char text_buf[TEXT_WIDTH+1];
    int i, length;

    struct card *card;

    assert (st->cur_card && st->first_index && st->last_index);

    bkgdset (COLOR_PAIR (BKGD_COLOR) | BKGD_ATTR | ' ');
    erase ();

/*attrset (COLOR_PAIR (CARD_COLOR) | CARD_ATTR);
move (0,0);
printw ("%d, %d", st->screen_cols, st->screen_rows);*/

    /* Drawing the title, card, and lines

       (a,p)--------(b,q)
         |            |
       (c,r)--------(d,s)
         |            |
         |            |
       (e,t)--------(f,u) */


    /* Calculate the corner coordinates */

    height = 1 + 1 + 1 + TEXT_HEIGHT + 1;

    a = c = e = 0;
    b = d = f = 1 + max (TITLE_WIDTH, TEXT_WIDTH)-1 + 1;
    p = q = (st->screen_rows - height) / 2;
    r = s = p + 1 + 1;
    t = u = r + TEXT_HEIGHT + 1;

    /* Draw the lines */

    attrset (COLOR_PAIR (CARD_COLOR) | CARD_ATTR);
    put_at (a, p, ACS_ULCORNER);
    print_between (a, p, b, q, ACS_HLINE);
    put_at (b, q, ACS_URCORNER);
    print_between (a, p, c, r, ACS_VLINE);
    print_between (b, q, d, s, ACS_VLINE);
    put_at (c, r, ACS_LTEE);
    print_between (c, r, d, s, ACS_HLINE);
    put_at (d, s, ACS_RTEE);
    print_between (c, r, e, t, ACS_VLINE);
    print_between (d, s, f, u, ACS_VLINE);
    put_at (e, t, ACS_LLCORNER);
    print_between (e, t, f, u, ACS_HLINE);
    put_at (f, u, ACS_LRCORNER);

    /* Draw the title

       +-------------+
       |(x,y)        |
       +-------------+
       |             |
       |             |
       +-------------+ */

    x = a + 1;
    y = p + 1;

    attrset (COLOR_PAIR (TEXT_COLOR) | TEXT_ATTR);
    memset (title_buf, ' ', TITLE_WIDTH);
    get_title (st->cur_card, title_buf, &length);
    print_at (x, y, title_buf, TITLE_WIDTH);

    /* Draw the text

       +-------------+
       |             |
       +-------------+
       |(x,y)        |
       |             |
       +-------------+ */

    x = c + 1;
    y = r + 1;

    for (i = 0; i < TEXT_HEIGHT; i++)
    {
        memset (text_buf, ' ', TEXT_WIDTH);
        get_row (st->cur_card, text_buf, &length, i);
        chop_text_linefeed (text_buf, length, ' ');
        print_at (x, y + i, text_buf, TEXT_WIDTH);
    }

    /* Draw the titles

            (x,y)
       +------+
       +------+
       +------+

                       */

    x = b + 2;
    y = 1;

    card = st->first_index;
    do
    {
        if (card == st->cur_card)
            attrset (COLOR_PAIR (HILITE_COLOR) | HILITE_ATTR);
        else
            attrset (COLOR_PAIR (INDEX_COLOR) | INDEX_ATTR);

        memset (title_buf, ' ', TITLE_WIDTH);
        get_title (card, title_buf, &length);
        print_at (x, y, title_buf, TITLE_WIDTH);

        card = card->next;
        y++;
    }
    while (card != st->last_index->next);

    /* Status */

    attrset (COLOR_PAIR (HILITE_COLOR) | HILITE_ATTR);
    switch (st->status)
    {
    case HELP:
        print_at_c (0, st->screen_rows-1,
        "Ctrl- eXit writeOut Dupe New Forward Back Kill Merge");
        break;
    case SKIPPED:
        print_at_c (0, st->screen_rows-1,
        "Warning: Images or OLE objects discarded");
        break;
    case LOADED:
        {
            char buffer[200];
            sprintf (buffer, "Loaded %d cards", count_cards (st->cur_card));
            print_at_c (0, st->screen_rows-1, buffer);
            break;
        }
    case SAVED:
        {
            char buffer[200];
            sprintf (buffer, "Saved %d cards", count_cards (st->cur_card));
            print_at_c (0, st->screen_rows-1, buffer);
            break;
        }
    }

    /* Position cursor */

    if (! st->title_mode)
    {
        x = c + 1;
        y = r + 1;

        x += real_x (st);
        y += real_y (st);
        move (y, x);
    }
    else
    {
        x = a + 1;
        y = p + 1;

        x += st->title_col;
        move (y, x);
    }

    refresh ();
}

void state_save (struct state *s)
{
    assert (s->cur_card && s->file);
    write_cardfile (s->file, s->cur_card);
    s->status = SAVED;
    state_refresh (s);
}

void state_delete_prev (struct state *s)
{
    if (! s->title_mode)
    {
        state_move_cursor (s, STAY);
        delete_prev (s->cur_card, &s->cursor_col, &s->cursor_row);
    }
    else
    {
        delete_title_prev (s->cur_card, &s->title_col);
        sort_cardfile (s->cur_card);
    }
    state_refresh (s);
}

void state_delete_at (struct state *s)
{
    if (! s->title_mode)
    {
        state_move_cursor (s, STAY);
        delete_at (s->cur_card, s->cursor_col, s->cursor_row);
    }
    else
    {
        delete_title_at (s->cur_card, s->title_col);
        sort_cardfile (s->cur_card);
    }
    state_refresh (s);
}

void state_free (struct state *s)
{
    assert (s->cur_card && s->file);
    delete_cardfile (s->cur_card);
    fclose (s->file);
    free (s);
}

void state_inform_width (struct state *s, int w)
{
    s->screen_cols = w;
}

void state_inform_height (struct state *s, int h)
{
    s->screen_rows = h;
}

void state_status_update (struct state *s)
{
   if (! s) return;

    switch (s->status)
    {
    case HELP: break;
    case SKIPPED:
    case LOADED:
    case SAVED: s->status = HELP; break;
    }
}


